@@ -14,7 +14,7 @@ pub use allocator::InstanceAllocator;
 pub use r#ref::{InstanceRef, WeakInstanceRef, WeakOrStrongInstanceRef};
 
 use crate::export::VMExtern;
-use crate::func_data_registry::{FuncDataRegistry, VMFuncRef};
+use crate::func_data_registry::VMFuncRef;
 use crate::global::Global;
 use crate::imports::Imports;
 use crate::memory::{Memory, MemoryError};
@@ -44,9 +44,9 @@ use std::slice;
 use std::sync::Arc;
 use wasmer_types::entity::{packed_option::ReservedValue, BoxedSlice, EntityRef, PrimaryMap};
 use wasmer_types::{
-    DataIndex, DataInitializer, ElemIndex, ExportIndex, FunctionIndex, GlobalIndex, GlobalInit,
-    LocalFunctionIndex, LocalGlobalIndex, LocalMemoryIndex, LocalTableIndex, MemoryIndex,
-    ModuleInfo, Pages, SignatureIndex, TableIndex, TableInitializer,
+    DataIndex, DataInitializer, ElemIndex, ExportIndex, FastGasCounter, FunctionIndex, GlobalIndex,
+    GlobalInit, InstanceConfig, LocalFunctionIndex, LocalGlobalIndex, LocalMemoryIndex,
+    LocalTableIndex, MemoryIndex, ModuleInfo, Pages, SignatureIndex, TableIndex, TableInitializer,
 };
 
 /// The function pointer to call with data and an [`Instance`] pointer to
@@ -66,6 +66,9 @@ pub(crate) struct Instance {
     /// The `ModuleInfo` this `Instance` was instantiated from.
     module: Arc<ModuleInfo>,
 
+    /// External configuration for instance.
+    config: InstanceConfig,
+
     /// Offsets in the `vmctx` region.
     offsets: VMOffsets,
 
@@ -93,8 +96,8 @@ pub(crate) struct Instance {
     /// get removed. A missing entry is considered equivalent to an empty slice.
     passive_data: RefCell<HashMap<DataIndex, Arc<[u8]>>>,
 
-    /// mapping of function indices to their func ref backing data.
-    funcrefs: BoxedSlice<FunctionIndex, VMFuncRef>,
+    /// TODO: document this
+    funcref_backings: BoxedSlice<FunctionIndex, Box<VMCallerCheckedAnyfunc>>,
 
     /// Hosts can store arbitrary per-instance information here.
     host_state: Box<dyn Any>,
@@ -398,6 +401,11 @@ impl Instance {
         unsafe { self.vmctx_plus_offset(self.offsets.vmctx_trap_handler()) }
     }
 
+    /// Return a pointer to the gas limiter.
+    pub fn gas_counter_ptr(&self) -> *mut *const FastGasCounter {
+        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_gas_limiter_pointer()) }
+    }
+
     /// Invoke the WebAssembly start function of the instance, if one is present.
     fn invoke_start_function(&self, trap_handler: &dyn TrapHandler) -> Result<(), Trap> {
         let start_index = match self.module.start_function {
@@ -640,7 +648,8 @@ impl Instance {
         if index == FunctionIndex::reserved_value() {
             return VMFuncRef::null();
         }
-        self.funcrefs[index]
+        let entry: &VMCallerCheckedAnyfunc = &*self.funcref_backings[index];
+        VMFuncRef(entry as *const VMCallerCheckedAnyfunc)
     }
 
     /// The `table.init` operation: initializes a portion of a table with a
@@ -909,9 +918,9 @@ impl InstanceHandle {
         finished_globals: BoxedSlice<LocalGlobalIndex, Arc<Global>>,
         imports: Imports,
         vmshared_signatures: BoxedSlice<SignatureIndex, VMSharedSignatureIndex>,
-        func_data_registry: &FuncDataRegistry,
         host_state: Box<dyn Any>,
         imported_function_envs: BoxedSlice<FunctionIndex, ImportFunctionEnv>,
+        instance_config: InstanceConfig,
     ) -> Result<Self, Trap> {
         let vmctx_globals = finished_globals
             .values()
@@ -923,10 +932,11 @@ impl InstanceHandle {
         let handle = {
             let offsets = allocator.offsets().clone();
             // use dummy value to create an instance so we can get the vmctx pointer
-            let funcrefs = PrimaryMap::new().into_boxed_slice();
+            let funcref_backings = PrimaryMap::new().into_boxed_slice();
             // Create the `Instance`. The unique, the One.
             let instance = Instance {
                 module,
+                config: instance_config.clone(),
                 offsets,
                 memories: finished_memories,
                 tables: finished_tables,
@@ -936,7 +946,7 @@ impl InstanceHandle {
                 passive_elements: Default::default(),
                 passive_data,
                 host_state,
-                funcrefs,
+                funcref_backings,
                 imported_function_envs,
                 vmctx: VMContext {},
             };
@@ -947,15 +957,15 @@ impl InstanceHandle {
             {
                 let instance = instance_ref.as_mut().unwrap();
                 let vmctx_ptr = instance.vmctx_ptr();
-                instance.funcrefs = build_funcrefs(
+                instance.funcref_backings = build_funcrefs(
                     &*instance.module,
                     &imports,
                     &instance.functions,
-                    func_data_registry,
                     &vmshared_signatures,
                     vmctx_ptr,
                 );
                 *(instance.trap_catcher_ptr()) = get_trap_handler();
+                *(instance.gas_counter_ptr()) = instance_config.gas_counter;
             }
 
             Self {
@@ -1245,7 +1255,6 @@ impl InstanceHandle {
         instance_ptr: *const ffi::c_void,
     ) -> Result<(), Err> {
         let instance_ref = self.instance.as_mut_unchecked();
-
         for import_function_env in instance_ref.imported_function_envs.values_mut() {
             match import_function_env {
                 ImportFunctionEnv::Env {
@@ -1443,10 +1452,9 @@ fn build_funcrefs(
     module_info: &ModuleInfo,
     imports: &Imports,
     finished_functions: &BoxedSlice<LocalFunctionIndex, FunctionBodyPtr>,
-    func_data_registry: &FuncDataRegistry,
     vmshared_signatures: &BoxedSlice<SignatureIndex, VMSharedSignatureIndex>,
     vmctx_ptr: *mut VMContext,
-) -> BoxedSlice<FunctionIndex, VMFuncRef> {
+) -> BoxedSlice<FunctionIndex, Box<VMCallerCheckedAnyfunc>> {
     let mut func_refs = PrimaryMap::with_capacity(module_info.functions.len());
 
     // do imported functions
@@ -1458,8 +1466,7 @@ fn build_funcrefs(
             type_index,
             vmctx: import.environment,
         };
-        let func_ref = func_data_registry.register(anyfunc);
-        func_refs.push(func_ref);
+        func_refs.push(Box::new(anyfunc));
     }
 
     // do local functions
@@ -1472,8 +1479,7 @@ fn build_funcrefs(
             type_index,
             vmctx: VMFunctionEnvironment { vmctx: vmctx_ptr },
         };
-        let func_ref = func_data_registry.register(anyfunc);
-        func_refs.push(func_ref);
+        func_refs.push(Box::new(anyfunc));
     }
 
     func_refs.into_boxed_slice()
